
class PullRequest(object):

    def __init__(self, number, url, head_ref, head_sha, base_ref, base_sha):
        self.number = number
        self.url = url
        self.head_ref = head_ref
        self.head_sha = head_sha
        self.base_ref = base_ref
        self.base_sha = base_sha



from __future__ import absolute_import

import functools
import json
import logging
import requests

from github import GithubException, UnknownObjectException, Github

from lintly.constants import LINTLY_IDENTIFIER
from lintly.formatters import build_pr_review_line_comment

from .base import BaseGitBackend
from .errors import NotFoundError, GitClientError
from .objects import PullRequest


logger = logging.getLogger(__name__)

# Get 100 items at a time so that we can make fewer API requests
DEFAULT_PER_PAGE = 100

GITHUB_API_HEADER = 'application/vnd.github.v3+json'
GITHUB_API_PR_REVIEW_HEADER = 'application/vnd.github.black-cat-preview+json'
GITHUB_DIFF_HEADER = 'application/vnd.github.3.diff'
GITHUB_USER_AGENT = 'Lintly'


def translate_github_exception(func):
    """
    Decorator to catch GitHub-specific exceptions and raise them as GitClientError exceptions.
    """

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except UnknownObjectException as e:
            logger.exception('GitHub API 404 Exception')
            raise NotFoundError(str(e))
        except GithubException as e:
            logger.exception('GitHub API Exception')
            raise GitClientError(str(e))

    return _wrapper


class GitHubAPIClient:
    """
    A wrapper class for making calls directly to the GitHub API and returning the results
    as JSON or a string (depending on the Content-Type header).
    """

    base_url = 'https://api.github.com'

    def __init__(self, token=None):
        self.token = token

    def get_headers(self):
        headers = {
            'Accept': GITHUB_API_HEADER,
            'Authorization': 'token {}'.format(self.token),
            'User-Agent': GITHUB_USER_AGENT,
        }
        return headers

    def post(self, url, data=None, headers=None):
        return self._do_request('post', url, json.dumps(data), headers)

    def get(self, url, data=None, headers=None):
        return self._do_request('get', url, data, headers)

    def put(self, url, data=None, headers=None):
        return self._do_request('put', url, json.dumps(data), headers)

    def _do_request(self, method, url, data=None, extra_headers=None):
        if data is None:
            data = dict()
        if extra_headers is None:
            extra_headers = dict()

        full_url = self.base_url + url
        headers = self.get_headers()
        headers.update(extra_headers)

        response = getattr(requests, method.lower())(full_url, data=data, headers=headers)
        if 200 <= response.status_code < 300:
            if 'application/json' in response.headers['Content-Type']:
                return response.json()
            else:
                return response.content
        elif response.status_code == 404:
            raise NotFoundError(response.content, status_code=response.status_code)
        else:
            raise GitClientError(response.content, status_code=response.status_code)


class GitHubBackend(BaseGitBackend):

    supports_pr_reviews = True

    def __init__(self, token, project, context):
        super(GitHubBackend, self).__init__(token, project)
        self.client = Github(token, user_agent=GITHUB_USER_AGENT, per_page=DEFAULT_PER_PAGE)
        self.context = context

    def _should_delete_comment(self, comment):
        return LINTLY_IDENTIFIER in comment.body

    @translate_github_exception
    def get_pull_request(self, pr):
        repo = self.client.get_repo(self.project.full_name)
        gh_pull = repo.get_pull(int(pr))

        pull_request = PullRequest(
            number=gh_pull.number,
            url=gh_pull.url,
            head_ref=gh_pull.head.ref,
            head_sha=gh_pull.head.sha,
            base_ref=gh_pull.base.ref,
            base_sha=gh_pull.base.sha
        )

        return pull_request

    @translate_github_exception
    def create_pull_request_comment(self, pr, comment):
        repo = self.client.get_repo(self.project.full_name)
        pull_request = repo.get_pull(int(pr))

        pull_request.create_issue_comment(
            body=comment
        )

    @translate_github_exception
    def delete_pull_request_comments(self, pr, bot):
        repo = self.client.get_repo(self.project.full_name)
        pull_request = repo.get_issue(int(pr))
        for comment in pull_request.get_comments():
            if self._should_delete_comment(comment):
                comment.delete()

    def get_pr_diff(self, pr):
        client = GitHubAPIClient(token=self.token)
        client.base_url = 'https://api.github.com'
        diff_url = '/repos/{owner}/{repo_name}/pulls/{pr_number}'.format(
            owner=self.project.owner_login,
            repo_name=self.project.name,
            pr_number=pr
        )

        diff = client.get(diff_url, headers={'Accept': GITHUB_DIFF_HEADER})

        return diff.decode('utf-8')

    def create_pull_request_review(self, pr, patch, all_violations):
        comments = []
        for file_path in all_violations:
            violations = all_violations[file_path]

            # https://developer.github.com/v3/pulls/comments/#input
            for violation in violations:
                patch_position = patch.get_patch_position(file_path, violation.line)
                if patch_position is not None:
                    comments.append({
                        'path': file_path,
                        'position': patch_position,
                        'body': build_pr_review_line_comment(violation)
                    })

        client = GitHubAPIClient(token=self.token)
        data = {
            'body': 'Lintly has detected code quality issues in this pull request.',
            'event': 'COMMENT',
            'comments': comments,
        }

        url = '/repos/{owner}/{repo_name}/pulls/{pr_number}/reviews'.format(
            owner=self.project.owner_login,
            repo_name=self.project.name,
            pr_number=pr
        )
        client.post(url, data, headers={'Accept': GITHUB_API_PR_REVIEW_HEADER})

    @translate_github_exception
    def delete_pull_request_review_comments(self, pr):
        repo = self.client.get_repo(self.project.full_name)
        pull_request = repo.get_pull(int(pr))
        for comment in pull_request.get_review_comments():
            if self._should_delete_comment(comment):
                comment.delete()

    def post_status(self, state, description, sha, target_url=''):
        url = '/repos/{owner}/{repo_name}/statuses/{sha}'.format(
            owner=self.project.owner_login, repo_name=self.project.name, sha=sha)

        # Using wrapper client since PyGitHub makes unnecessary API calls
        client = GitHubAPIClient(token=self.token)
        data = {
            'state': state,
            'description': description,
            'target_url': target_url,
            'context': self.context
        }
        client.post(url, data)



import logging
import logging.config
import sys

import click

from .builds import LintlyBuild
from .config import Config
from .constants import FAIL_ON_ANY, FAIL_ON_NEW
from .exceptions import NotPullRequestException
from .parsers import PARSERS


logger = logging.getLogger(__name__)


@click.command()
@click.option('--api-key',
              envvar='LINTLY_API_KEY',
              help='The GitHub API key to use for commenting on PRs')
@click.option('--repo',
              envvar='LINTLY_REPO',
              help='The GitHub repo name in the format {owner}/{repo}')
@click.option('--pr',
              envvar='LINTLY_PR',
              help='The pull request number for this build')
@click.option('--commit-sha',
              envvar='LINTLY_COMMIT_SHA',
              help='The commit Lintly is running against.')
@click.option('--format',
              envvar='LINTLY_FORMAT',
              type=click.Choice(list(PARSERS.keys())),
              default='unix',
              help='The linting output format Lintly should expect to receive')
@click.option('--context',
              help='Override the commit status context')
@click.option('--fail-on',
              envvar='LINTLY_FAIL_ON',
              type=click.Choice([FAIL_ON_ANY, FAIL_ON_NEW]),
              default=FAIL_ON_ANY,
              help='Whether Lintly should fail if any violations are detected or only if new violations are detected.')
@click.option('--post-status/--no-post-status',
              default=True,
              help='Used to determine if Lintly should post a PR status to GitHub.')
@click.option('--log',
              is_flag=True,
              help='Send Lintly debug logs to the console.')
@click.option('--exit-zero/--no-exit-zero', default=False,
              help='Whether Lintly should exit with error code indicating amount of violations or not')
def main(**options):
    """Slurp up linter output and send it to a GitHub PR review."""
    configure_logging(log_all=options.get('log'))

    stdin_stream = click.get_text_stream('stdin')
    stdin_text = stdin_stream.read()

    click.echo(stdin_text)

    config = Config(options)

    build = LintlyBuild(config, stdin_text)
    try:
        build.execute()
    except NotPullRequestException:
        logger.info('Not a PR. Lintly is exiting.')
        sys.exit(0)

    exit_code = 0
    # Exit with the number of files that have violations
    if not options['exit_zero']:
        exit_code = len(build.violations)
    sys.exit(exit_code)


def configure_logging(log_all=False):
    log_level = 'DEBUG' if log_all else 'WARNING'
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'standard': {
                'format': 'Lintly: %(asctime)s [%(levelname)s] %(name)s: %(message)s'
            },
        },
        'handlers': {
            'default': {
                'level': log_level,
                'class': 'logging.StreamHandler',
                'formatter': 'standard'
            },
        },
        'loggers': {
            'lintly': {
                'handlers': ['default'],
                'level': log_level,
                'propagate': True
            }
        }
    })




class BaseGitBackend(object):
    """
    A base class for making calls to external Git APIs.
    """

    supports_pr_reviews = False

    def __init__(self, token, project):
        self.token = token
        self.project = project

    def __repr__(self):
        token = '********' if self.token else 'None'
        return '{}(token={}, project={})'.format(
            self.__class__.__name__, token, self.project)

    def get_pull_request(self, pr):
        """
        Returns a pull request for a given PR number.
        """
        raise NotImplementedError

    def create_pull_request_comment(self, pr, comment):
        """
        Creates a comment on a pull request.
        """
        raise NotImplementedError

    def delete_pull_request_comments(self, pr, bot):
        """
        Deletes all pull request comments for the bot account.
        """
        raise NotImplementedError

    def get_pr_diff(self, pr):
        """
        Returns a diff for a pull request.
        """
        raise NotImplementedError

    def create_pull_request_review(self, pr, patch, all_violations):
        """
        Creates a pull request review for the given build.
        """
        raise NotImplementedError

    def delete_pull_request_review_comments(self, pr):
        """
        Deletes all pull request review comments for the bot account.
        """
        raise NotImplementedError

    def post_status(self, state, description, sha, target_url):
        raise NotImplementedError




class GitClientError(Exception):
    """Base class for all exceptions while querying Git clients like GitHub or GitLab."""

    def __init__(self, message=None, status_code=None):
        super(GitClientError, self).__init__(message)
        self.status_code = status_code


class UnauthorizedError(GitClientError):
    pass


class NotFoundError(GitClientError):
    pass


class NotSupportedError(GitClientError):
    pass



class Violation(object):

    def __init__(self, line, column, code, message):
        self.line = line
        self.column = column
        self.code = code
        self.message = message

    def __str__(self):
        return self.message

    def __repr__(self):
        return 'Violation(line={}, column={}, code="{}", message="{}")'.format(
            self.line, self.column, self.code, self.message
        )



import ci

REDACTED = '********'


class Config(object):
    """A Config object that knows how to return configuration from the CLI or Continuous Integration services"""

    def __init__(self, cli_config):
        self.cli_config = cli_config

    def as_dict(self):
        return {
            'pr': self.pr,
            'repo': self.repo,
            'commit_sha': self.commit_sha,
            'api_key': REDACTED,
            'format': self.format,
            'fail_on': self.fail_on,
            'post_status': self.post_status,
        }

    @property
    def pr(self):
        return self.cli_config['pr'] or ci.pr()

    @property
    def repo(self):
        return self.cli_config['repo'] or ci.repo()

    @property
    def commit_sha(self):
        return self.cli_config['commit_sha'] or ci.commit_sha()

    @property
    def context(self):
        return self.cli_config['context']

    @property
    def api_key(self):
        return self.cli_config['api_key']

    @property
    def format(self):
        return self.cli_config['format']

    @property
    def fail_on(self):
        return self.cli_config['fail_on']

    @property
    def post_status(self):
        return self.cli_config['post_status']



import collections
import logging

from .constants import FAIL_ON_ANY
from .exceptions import NotPullRequestException
from .backends.github import GitHubBackend
from .backends.errors import GitClientError
from .formatters import build_pr_comment
from .parsers import PARSERS
from .patch import Patch
from .projects import Project


logger = logging.getLogger(__name__)


class LintlyBuild(object):

    def __init__(self, config, linter_output):
        self.config = config
        self.linter_output = linter_output

        self.project = Project(config.repo)

        context = config.context or "Lintly/{0}".format(config.format)
        self.git_client = GitHubBackend(token=config.api_key, project=self.project, context=context)

        # All violations found from the linting output
        self._all_violations = {}

        # Violations that are only caused by changes to the current PR
        self._diff_violations = {}

    @property
    def violations(self):
        """
        Returns either the diff violations or all violations depending on configuration.
        """
        return self._all_violations if self.config.fail_on == FAIL_ON_ANY else self._diff_violations

    @property
    def has_violations(self):
        return bool(self.violations)

    @property
    def introduced_issues_count(self):
        if self.violations is None:
            return 0
        return sum(len(item[1]) for item in self.violations.items())

    def execute(self):
        """
        Executes a new build on a project.
        """
        if not self.config.pr:
            raise NotPullRequestException

        logger.debug('Using the following configuration:')
        for name, value in self.config.as_dict().items():
            logger.debug('  - {}={}'.format(name, repr(value)))

        logger.info('Running Lintly against PR #{} for repo {}'.format(self.config.pr, self.project))

        parser = PARSERS.get(self.config.format)
        self._all_violations = parser.parse_violations(self.linter_output)
        logger.info('Lintly found violations in {} files'.format(len(self._all_violations)))

        diff = self.get_pr_diff()
        patch = self.get_pr_patch(diff)
        self._diff_violations = self.find_diff_violations(patch)
        logger.info('Lintly found diff violations in {} files'.format(len(self._diff_violations)))

        self.post_pr_comment(patch)
        self.post_commit_status()

    def get_pr_diff(self):
        return self.git_client.get_pr_diff(self.config.pr)

    def get_pr_patch(self, diff):
        return Patch(diff)

    def find_diff_violations(self, patch):
        """
        Uses the diff for this build to find changed lines that also have violations.
        """
        violations = collections.defaultdict(list)
        for line in patch.changed_lines:
            file_violations = self._all_violations.get(line['file_name'])
            if not file_violations:
                continue

            line_violations = [v for v in file_violations if v.line == line['line_number']]

            for v in line_violations:
                violations[line['file_name']].append(v)

        return violations

    def post_pr_comment(self, patch):
        """
        Posts a comment to the GitHub PR if the diff results have issues.
        """
        if self.has_violations:
            post_pr_comment = True

            # Attempt to post a PR review. If posting the PR review fails because the bot account
            # does not have permission to review the PR then simply revert to posting a regular PR
            # comment.
            try:
                logger.info('Deleting old PR review comments')
                self.git_client.delete_pull_request_review_comments(self.config.pr)

                logger.info('Creating PR review')
                self.git_client.create_pull_request_review(self.config.pr, patch, self._diff_violations)
                post_pr_comment = False
            except GitClientError as e:
                # TODO: Make `create_pull_request_review` raise an `UnauthorizedError`
                # so that we don't have to check for a specific message in the exception
                if 'Viewer does not have permission to review this pull request' in str(e):
                    logger.warning("Could not post PR review (the account didn't have permission)")
                    pass
                else:
                    raise

            if post_pr_comment:
                logger.info('Deleting old PR comment')
                self.git_client.delete_pull_request_comments(self.config.pr)

                logger.info('Creating PR comment')
                comment = build_pr_comment(self.config, self.violations)
                self.git_client.create_pull_request_comment(self.config.pr, comment)

    def post_commit_status(self):
        """
        Posts results to a commit status in GitHub if this build is for a pull request.
        """
        if self.violations:
            plural = '' if self.introduced_issues_count == 1 else 's'
            description = 'Pull Request introduced {} linting violation{}'.format(
                self.introduced_issues_count, plural)
            self._post_status('failure', description)
        else:
            self._post_status('success', 'Linting detected no new issues.')

    def _post_status(self, state, description):
        """
        Creates a GitHub status for this build's commit if enabled for the project.
        :param state: The state of the status (pending, error, failure, success)
        :param description: The description for the status
        """
        if self.config.post_status:
            logger.info('Commit statuses enabled')
            if self.config.commit_sha:
                logger.info('Posting {} status to commit SHA {}'.format(state, self.config.commit_sha))
                self.git_client.post_status(
                    state,
                    description,
                    sha=self.config.commit_sha
                )
            else:
                logger.warning('Cannot post commit status because no commit SHA has been '
                               'specified. Either use the --commit-sha CLI argument or '
                               'set the LINTLY_COMMIT_SHA environment variable.')
        else:
            logger.info('Commit statuses disabled')



# -*- coding: utf-8 -*-
"""
Parsers accept linter output and return file paths and all of the violations in that file.
"""
import collections
import json
import os
import re

from .violations import Violation


class BaseLintParser(object):

    def parse_violations(self, output):
        raise NotImplementedError

    def _get_working_dir(self):
        return os.getcwd()

    def _normalize_path(self, path):
        """
        Normalizes a file path so that it returns a path relative to the root repo directory.
        """
        norm_path = os.path.normpath(path)
        return os.path.relpath(norm_path, start=self._get_working_dir())


class LineRegexParser(BaseLintParser):
    """
    A parser that runs a regular expression on each line of the output to return violations.
    The regex should match the following capture groups:

        - path
        - line
        - column
        - code
        - message
    """

    def __init__(self, regex):
        self.regex = regex

    def parse_violations(self, output):
        violations = collections.defaultdict(list)

        line_regex = re.compile(self.regex)

        # Collect all the issues into a dict where the keys are the file paths and the values are a
        # list of the issues in that file.
        for line in output.strip().splitlines():
            clean_line = line.strip()

            match = line_regex.match(clean_line)

            if not match:
                continue

            path = self._normalize_path(match.group('path'))

            violation = Violation(
                line=int(match.group('line')),
                column=int(match.group('column')),
                code=match.group('code'),
                message=match.group('message')
            )

            violations[path].append(violation)

        return violations


class PylintJSONParser(BaseLintParser):
    """
    Pylint JSON format:

        [
            {
                "type": "convention",
                "module": "lintly.backends.base",
                "obj": "BaseGitBackend.post_status",
                "line": 54,
                "column": 4,
                "path": "lintly/backends/base.py",
                "symbol": "missing-docstring",
                "message": "Missing method docstring",
                "message-id": "C0111"
            }
        ]
    """

    def parse_violations(self, output):
        # Sometimes pylint will output "No config file found, using default configuration".
        # This handles that case by removing that line.
        if output and output.startswith('No config'):
            output = '\n'.join(output.splitlines()[1:])

        output = output.strip()
        if not output:
            return dict()

        json_data = json.loads(output)

        violations = collections.defaultdict(list)

        for violation_json in json_data:
            violation = Violation(
                line=violation_json['line'],
                column=violation_json['column'],
                code='{} ({})'.format(violation_json['message-id'], violation_json['symbol']),
                message=violation_json['message']
            )

            path = self._normalize_path(violation_json['path'])
            violations[path].append(violation)

        return violations


class ESLintParser(BaseLintParser):

    def parse_violations(self, output):
        violations = collections.defaultdict(list)

        current_file = None

        # Collect all the issues into a dict where the keys are the file paths and the values are a
        # list of the issues in that file.
        for line in output.strip().splitlines():
            if line.startswith(' '):
                # This line is a linting violation
                regex = r'^(?P<line>\d+):(?P<column>\d+)\s+(error|warning)\s+(?P<message>.*)\s+(?P<code>.+)$'
                match = re.match(regex, line.strip())

                violation = Violation(
                    line=int(match.group('line')),
                    column=int(match.group('column')),
                    code=match.group('code').strip(),
                    message=match.group('message').strip()
                )
                violations[current_file].append(violation)
            elif line.startswith('✖'):
                # We're at the end of the file
                break
            else:
                # This line is a file path
                current_file = self._normalize_path(line)

        return violations


class StylelintParser(BaseLintParser):

    def parse_violations(self, output):
        violations = collections.defaultdict(list)

        current_file = None

        # Collect all the issues into a dict where the keys are the file paths and the values are a
        # list of the issues in that file.
        for line in output.strip().splitlines():
            if line.startswith(' '):
                # This line is a linting violation
                regex = r'^(?P<line>\d+):(?P<column>\d+)\s+✖\s+(?P<message>.*)\s+(?P<code>.+)$'
                match = re.match(regex, line.strip())

                violation = Violation(
                    line=int(match.group('line')),
                    column=int(match.group('column')),
                    code=match.group('code').strip(),
                    message=match.group('message').strip()
                )
                violations[current_file].append(violation)
            else:
                # This line is a file path
                current_file = self._normalize_path(line)

        return violations


class BlackParser(BaseLintParser):
    """A parser for the `black [source] --check` command."""

    def parse_violations(self, output):
        violations = {}
        for line in output.strip().splitlines():
            # That means a file needs to be formatted by `black`.
            if line.startswith('would reformat '):
                # Last part is the file path.
                path = self._normalize_path(line.split(' ')[-1])
                violations[path] = [Violation(
                    line=1, column=1, code='`black`', message='this file needs to be formatted'
                )]
        return violations


class CfnLintParser(BaseLintParser):
    """A parser for the `cfn-lint` command.

      cfn-lint output example:

      W2001 Parameter UnusedParameter not used.
      template.yaml:2:9
    """

    def parse_violations(self, output):
        violations = collections.defaultdict(list)

        regex = re.compile(r"[EW]\d{4}\s")

        next_line_is_path = False
        current_violation = None
        for line in output.strip().splitlines():
            if regex.match(line):
                # This line is a cfn-lint error or warning
                next_line_is_path = True
                current_violation = line
            elif next_line_is_path:
                # This line is a filepath:line_number:column_number
                path, line_number, column = line.split(":")
                path = self._normalize_path(path)

                code, message = current_violation.split(" ", 1)

                violation = Violation(line=int(line_number),
                                      column=int(column),
                                      code=code,
                                      message=message)
                violations[path].append(violation)

                next_line_is_path = False
                current_violation = None

        return violations


class CfnNagParser(BaseLintParser):

    def parse_violations(self, output):

        file_list = json.loads(output)
        violations = {}

        for file in file_list:
            file_violations = []
            for violation_info in file["file_results"]["violations"]:
                for line_number in violation_info["line_numbers"]:
                    violation = Violation(
                        line=line_number,
                        column=0,
                        code=violation_info["id"],
                        message=violation_info["message"]
                    )

                    file_violations.append(violation)

            violations[file["filename"]] = file_violations

        return violations


DEFAULT_PARSER = LineRegexParser(r'^(?P<path>.*):(?P<line>\d+):(?P<column>\d+): (?P<code>\w\d+) (?P<message>.*)$')


PARSERS = {
    # Default flake8 format
    # docs/conf.py:230:1: E265 block comment should start with '# '
    # path:line:column: CODE message
    'unix': DEFAULT_PARSER,
    'flake8': DEFAULT_PARSER,

    # Pylint ---output-format=json
    'pylint-json': PylintJSONParser(),

    # ESLint's default formatter
    # /Users/grant/project/file1.js
    #     1:1    error  '$' is not defined                              no-undef
    'eslint': ESLintParser(),

    # ESLint's unix formatter
    # lintly/static/js/scripts.js:69:1: 'lintly' is not defined. [Error/no-undef]
    # path:line:column: message [CODE]
    'eslint-unix': LineRegexParser(r'^(?P<path>.*):(?P<line>\d+):(?P<column>\d+): '
                                   r'(?P<message>.+) \[(Warning|Error)/(?P<code>.+)\]$'),

    # Stylelint's default formatter
    # lintly/static/sass/file1.scss
    #   13:1  ✖  Expected no more than 1 empty line   max-empty-lines
    'stylelint': StylelintParser(),

    # Black's check command default formatter.
    'black': BlackParser(),

    # cfn-lint default formatter
    'cfn-lint': CfnLintParser(),

    # cfn-nag JSON output
    'cfn-nag': CfnNagParser(),
}



from .base import BaseGitBackend


class DummyGitBackend(BaseGitBackend):
    """
    A Git backend class that returns dummy data for all of the abstract methods.
    """

    supports_pr_reviews = False

    def get_pull_request(self, pr):
        pass

    def create_pull_request_comment(self, pr, comment):
        pass

    def delete_pull_request_comments(self, pr, bot):
        pass

    def create_pull_request_review(self, pr, patch, all_violations):
        pass

    def delete_pull_request_review_comments(self, pr):
        pass

    def post_status(self, state, description, sha, target_url):
        pass



FAIL_ON_ANY = 'any'
FAIL_ON_NEW = 'new'

# Identifies that a comment came from Lintly. This is used to aid in automatically
# deleting old PR comments/reviews. This is valid Markdown that is hidden from
# users in GitHub and GitLab.
LINTLY_IDENTIFIER = '<!-- Automatically posted by Lintly -->'



import logging
import re

from cached_property import cached_property


FILE_NAME_LINE = re.compile(r'^\+\+\+ b/(?P<file_name>.+)')
RANGE_INFORMATION_LINE = re.compile(r'^@@ .+\+(?P<line_number>\d+),')
MODIFIED_LINE = re.compile(r'^\+(?!\+|\+)')
NOT_REMOVED_OR_NEWLINE_WARNING = re.compile(r'^[^-\\]')


logger = logging.getLogger(__name__)


class Patch(object):
    """
    Parses the body of a diff and returns the lines that changed as well as their "position",
    as outlined by GitHub here: https://developer.github.com/v3/pulls/comments/#create-a-comment
    """

    def __init__(self, body=''):
        self.body = body

    @cached_property
    def changed_lines(self):
        """
        A list of dicts in the format:
            {
                'file_name': str,
                'content': str,
                'line_number': int,
                'position': int
            }
        """
        lines = []
        file_name = ''
        line_number = 0
        patch_position = -1
        found_first_information_line = False

        for i, content in enumerate(self.body.splitlines()):
            range_information_match = RANGE_INFORMATION_LINE.search(content)
            file_name_line_match = FILE_NAME_LINE.search(content)

            if file_name_line_match:
                file_name = file_name_line_match.group('file_name')
                found_first_information_line = False
            elif range_information_match:
                line_number = int(range_information_match.group('line_number'))
                if not found_first_information_line:
                    # This is the first information line. Set patch position to 1 and start counting
                    patch_position = 0
                    found_first_information_line = True
            elif MODIFIED_LINE.search(content):
                line = {
                    'file_name': file_name,
                    'content': content,
                    'line_number': line_number,
                    'position': patch_position
                }

                lines.append(line)
                line_number += 1
            elif NOT_REMOVED_OR_NEWLINE_WARNING.search(content) or content == '':
                line_number += 1

            patch_position += 1

        return lines

    def get_patch_position(self, file_name, line_number):
        matching_lines = [line for line in self.changed_lines
                          if line['file_name'] == file_name and line['line_number'] == line_number]

        if len(matching_lines) == 0:
            return None
        elif len(matching_lines) == 1:
            return matching_lines[0]['position']
        else:
            logger.warning('Invalid patch or build.')
            logger.warning('Multiple matching lines found for {file_name} on line '
                           '{line_number}'.format(file_name=file_name, line_number=line_number))



"""
Slurp up linter output and send it to a GitHub PR review.
"""
import codecs
import os
from setuptools import find_packages, setup

dependencies = [
    'ci-py',
    'cached-property<2.0',
    'click<7.0',
    'jinja2<3.0',
    'PyGithub<2.0',
    'python-gitlab<2.0',
    'six',
]


def read(*parts):
    filename = os.path.join(os.path.dirname(__file__), *parts)
    with codecs.open(filename, encoding='utf-8') as fp:
        return fp.read()


setup(
    name='lintly',
    version='0.4.2',
    url='https://github.com/grantmcconnaughey/lintly',
    license='MIT',
    author='Grant McConnaughey',
    author_email='grantmcconnaughey@gmail.com',
    description='Slurp up linter output and send it to a GitHub PR review.',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=dependencies,
    entry_points={
        'console_scripts': [
            'lintly = lintly.cli:main',
        ],
    },
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Quality Assurance',
    ]
)



class Project(object):

    def __init__(self, full_name):
        self.full_name = full_name

    def __str__(self):
        return self.full_name

    @property
    def owner_login(self):
        return self.full_name.split('/')[0]

    @property
    def name(self):
        return self.full_name.split('/')[1]



from __future__ import absolute_import

import functools
import json
import logging

import gitlab
import requests

from lintly.constants import LINTLY_IDENTIFIER

from .base import BaseGitBackend
from .errors import (
    NotSupportedError, NotFoundError, GitClientError, UnauthorizedError
)
from .objects import PullRequest


logger = logging.getLogger(__name__)

# Get 100 items at a time so that we can make fewer API requests
DEFAULT_PER_PAGE = 100

GITLAB_URL = 'https://gitlab.com'

GITLAB_API_VERSION = 4


def translate_gitlab_exception(func):
    """
    Decorator to catch GitLab-specific exceptions and raise them as GitClientError exceptions.
    """

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except gitlab.GitlabError as e:
            status_to_exception = {
                401: UnauthorizedError,
                404: NotFoundError,
            }

            exc_class = status_to_exception.get(e.response_code, GitClientError)
            raise exc_class(str(e), status_code=e.response_code)

    return _wrapper


class GitLabAPIClient:
    """
    A wrapper class for making calls directly to the GitHub API and returning the results
    as JSON or a string (depending on the Content-Type header).
    """

    base_url = '{url}/api/v{version}'.format(url=GITLAB_URL, version=GITLAB_API_VERSION)

    def __init__(self, token=None, project=None):
        self.project = project
        self.token = token

    def __repr__(self):
        token = 'REDACTED' if self.token else 'None'
        return 'GitLabAPIClient(token={}, user={}, project={})'.format(
            token, self.user, self.project)

    def get_headers(self):
        headers = {
            'Authorization': 'Bearer {}'.format(self.token),
        }
        return headers

    def post(self, url, data=None, headers=None):
        return self._do_request('post', url, json.dumps(data), headers)

    def get(self, url, data=None, headers=None):
        return self._do_request('get', url, data, headers)

    def delete(self, url, data=None, headers=None):
        return self._do_request('delete', url, json.dumps(data), headers)

    def _do_request(self, method, url, data=None, extra_headers=None):
        if data is None:
            data = dict()
        if extra_headers is None:
            extra_headers = dict()

        full_url = self.base_url + url
        headers = self.get_headers()
        headers.update(extra_headers)

        response = getattr(requests, method.lower())(full_url, data=data, headers=headers)
        if 200 <= response.status_code < 300:
            if 'application/json' in response.headers['Content-Type']:
                return response.json()
            else:
                return response.content
        elif response.status_code == 404:
            raise NotFoundError(response.content, status_code=response.status_code)
        else:
            raise GitClientError(response.content, status_code=response.status_code)


class GitLabBackend(BaseGitBackend):

    supports_pr_reviews = False

    def __init__(self, token, project):
        super(GitLabBackend, self).__init__(token, project)
        self.client = gitlab.Gitlab(GITLAB_URL, token, api_version=str(GITLAB_API_VERSION))

    @translate_gitlab_exception
    def get_pull_request(self, pr):
        project = self.client.projects.get(self.project.full_name)
        mr = project.mergerequests.list(iid=pr)[0]
        target_sha = self.get_latest_branch_commit_sha(mr.target_branch)
        pull_request = PullRequest(
            number=pr,
            url=mr.web_url,
            head_ref=mr.source_branch,
            head_sha=mr.sha,
            base_ref=mr.target_branch,
            base_sha=target_sha
        )
        return pull_request

    @translate_gitlab_exception
    def create_pull_request_comment(self, pr, comment):
        project = self.client.projects.get(self.project.full_name)
        mr = project.mergerequests.list(iid=pr)[0]
        mr.notes.create({'body': comment})

    @translate_gitlab_exception
    def delete_pull_request_comments(self, pr, bot):
        project = self.client.projects.get(self.project.full_name)
        mr = project.mergerequests.list(iid=pr)[0]
        client = GitLabAPIClient(self.token, self.user, self.project)
        for note in mr.notes.list(all=True, per_page=DEFAULT_PER_PAGE):
            if LINTLY_IDENTIFIER in note.body:
                url = '/projects/{project_id}/merge_requests/{mr_id}/notes/{note_id}'.format(
                    project_id=project.id, mr_id=mr.id, note_id=note.id
                )
                client.delete(url)

    @translate_gitlab_exception
    def create_pull_request_review(self, pr, patch, all_violations):
        raise NotSupportedError()

    @translate_gitlab_exception
    def delete_pull_request_review_comments(self, pr):
        raise NotSupportedError()

    @translate_gitlab_exception
    def post_status(self, state, description, sha, target_url):
        # TODO: Fix this ugliness...
        if state == 'failure':
            state = 'failed'
        project = self.client.projects.get(self.project.full_name)
        gl_commit = project.commits.get(sha)
        gl_commit.statuses.create({'state': state,
                                   'description': description,
                                   'target_url': target_url,
                                   'name': 'Lintly'})



"""
Formats text that will be posted to Pull Requests.
"""
import os

from jinja2 import Environment, FileSystemLoader

from .constants import LINTLY_IDENTIFIER


TEMPLATES_PATH = os.path.join(os.path.dirname(__file__), 'templates')


env = Environment(
    loader=FileSystemLoader(TEMPLATES_PATH),
    autoescape=False
)


def build_pr_comment(config, violations):
    """
    Creates a Markdown representation of the comment to be posted to a pull request.
    :return: The comment
    """
    template = env.get_template('pr_comment.txt')
    return template.render(violations=violations, LINTLY_IDENTIFIER=LINTLY_IDENTIFIER)


def build_pr_review_line_comment(violation):
    """
    Creates a Markdown representation of the comment to be posted to a pull request.
    :return: The comment
    """
    template = env.get_template('pr_review_line_comment.txt')
    return template.render(violation=violation, LINTLY_IDENTIFIER=LINTLY_IDENTIFIER)
